#include <string>     // std::string, std::stoi
#include <iostream>   // std::cout
#include <vector>
#include <tuple>
#include <unistd.h> //file access
#include <random>
#include <fstream>
#include <set>

using namespace std;

typedef pair<unsigned int, unsigned int>  interval;


/**
* @brief random number generator 
*/
struct my_lkg{
private: 
	minstd_rand0 lkg;

public:
	my_lkg(unsigned int seed = 0){
		lkg = minstd_rand0(seed);
	}
	
	//random ui
	unsigned int get_ui(){
		if (lkg()%2) 
			return lkg();
		else
			return lkg() + 2147483648;
	}
	
	//random ui
	unsigned int operator()(){
		return get_ui();
	}
	
	//random ui in range 1 ... max
	unsigned int operator()(unsigned int max){
		return get_ui() % max + 1;
	}
	
	//random ui in range I.first ... I.second
	unsigned int operator()(interval I){
		return get_ui() % (I.second - I.first + 1) + I.first;
	}
};

/**
* @brief random base generator 
*/
struct random_base{
private:
	my_lkg lkg;
public:
	random_base(unsigned int seed = 0){
		lkg = my_lkg(seed);
	}
	
	char operator()(){
		switch(lkg(4)){
			case 1: return 'A';
			case 2: return 'T';
			case 3: return 'G';
			default: return 'C';
		}
	}
};


template<class T, class A = allocator<T>>
/**
* @brief prints a vector for debugging
* 
* @param T vector type
* @param A vector alloc type
* @param v vector
*/
void print_vector(vector<T,A> const& v){
	for(unsigned int i = 0; i< v.size(); i++){
		cout << v[i] << " ";
	}
	cout << endl;
};

/**
* @brief prints help message
*/
void print_help(){
	cout << "This program generate DAGs." << endl
	<< "There are 2 diffrent methods for generation." << endl
	<< "A topical order of the nodes is always 1 < 2 < 3 ..." << endl<< endl
	<< "Version1:" << endl
	<< "You input the number of nodes N, paths P and the length of a path PL" << endl
	<< "The progam will generate a graph with can be covered with P paths (P is not the size of the minimum PC)"
	<< "Therefor the program will generate P paths"
	<< "The nodes of each path will be selected by randomly choosing a node and adding it to the path"
	<< "after adding PL nodes (duplicates possilbe), the edges of this path will be added to graph"
	<< "When the last path is generated, it will be ensured, that this path contains all nodes that were not part of any path before"<< endl<< endl
	<< "Version2:"<< endl
	<< "You input the number of nodes N and edges E" << endl
	<< "The program will add a edges from s to each node and from each node to t" << endl
	<< "Then E random edges (u,v) with u < v will added to the graph" << endl<< endl
	<< "Options:" << endl	
	<< "-n N, numer of nodes (N <= 4294967295)" << endl
	<< "-c count, numer of graphs to generate" << endl
	<< "-o out-dir, directory to write gfa files in (default ./)" << endl
	<< "-s seed, seed for the RNG (default 0)" << endl 
	<< "-v2, to use version 2 (default version 1)" << endl<< endl
	<< "Additional options for version 1:" << endl
	<< "-p P, numer of paths" << endl
	<< "-pl PL, numer of nodes per path (approximatly)" << endl<< endl
	<< "Additional options for version 2:" << endl	
	<< "-e E, numer of edges (approximatly)" << endl << endl	
	<< "For the variables N, E, P and PL you can set a interval of which the used value will be chosen randomly" << endl	
	<< "e_min, e_max, p_min, p_max, pl_min, pl_max , to set the boundaries of the interval" << endl;

}



/**
* @brief fills edges of a path to the edges set 
* @param path vector with v[i] = 1 iff i is part of path
* @param covered vector with v[i] = 1 iff i is part of any path
* @param edges set of edges
*/
void report_path(vector<bool> & path, vector<bool> & covered,set<unsigned long> & edges){
	
//print_vector(path);
//report path
	unsigned int u = 0;
	for(unsigned int j = 1; j < path.size()-1; j++){
		if(path[j]){
			covered[j] = true;
			unsigned long e = u;
			e = e << 32;
			e += j;
			edges.insert(e);
			u = j;
		}
	}
	unsigned long e = u;
	e = e << 32;
	e += path.size()-1;
	edges.insert(e);
};

/**
* @brief generate a random DAG, whith "path method"
* 
* @param N Number of nodes
* @param P_i Interval with number of paths
* @param PL_i Interval with legth of path
* @param s seed for rng
* @return edges and name-string
*/
tuple<set<unsigned long>,string> version_1(unsigned int N, interval P_i, interval PL_i, unsigned int s){
	vector<bool> covered (N,false);
	set<unsigned long> edges;
	my_lkg lkg = my_lkg(s); 
	
	unsigned int P = lkg(P_i);
	unsigned int PL = lkg(PL_i);
	
	for(unsigned int i = 1; i < P; i++ ){	//genereate P-1 Paths	
		//init empty path
		vector<bool> path (N,false);		
		//fill path
		for(unsigned int j = 0; j < PL; j++ )
			path[lkg(N-1)] = 1;		
		report_path(path,covered,edges);	//fill in edges of Path
	}
	
	
	//add another path that includes all missing nodes
	//init empty path
	vector<bool> path (N,false);
	unsigned int m = 0;//fill in nodes that were not covered	
	for(unsigned int j = 1; j < N-1; j++){
		if(!covered[j]){
			path[j] = true;
			m++;
		}
	} 
	//fill more nodes to path
	for(; m < PL; m++ ){
		path[lkg(N-1)] = 1;
	}	
	//report path
	report_path(path,covered,edges);
	
	string fname = "N" + to_string(N) + "_P" + to_string(P) + "_PL" + to_string(PL) + "_S" + to_string(s);
 	return make_tuple(edges,fname);
}

/**
* @brief generates DAG with Erdos Renyi model
* 
* @param N Number of nodes
* @param E_i Interval with number of edges
* @param s seed for rng
* @return edges and name-string
*/
tuple<set<unsigned long>,string> version_2(unsigned int N, interval E_i, unsigned int s){
	my_lkg lkg = my_lkg(s); 
	
	unsigned int E = lkg(E_i);
	
	set<unsigned long> edges;
	
	for(unsigned int i = 0; i < E; i++){
		unsigned int u = lkg(N-1);
		unsigned int v = lkg(N-1);
		
		if(u == v) {
			i--; 
			continue;
		}
		
		if(u > v) 
			swap(u,v);
		
		unsigned long e = u;
		e <<= 32;
		e += v;
		edges.insert(e);
	}
	
	for(unsigned int u = 1; u < N-1; u++){
		unsigned long e1 = 0;
		unsigned long e2 = u;
		e1 <<= 32;
		e2 <<= 32;
		e1 += u;
		e2 += N-1;
		edges.insert(e1);
		edges.insert(e2);
	}
	
	string fname = "N" + to_string(N) + "_E" + to_string(E)  + "_S" + to_string(s);
	return make_tuple(edges, fname);
	
}

void write_gfa_file(string fname, unsigned int N,  set<unsigned long> edges, unsigned int s){
	random_base rb = random_base(s);
	
	ofstream myfile;
	myfile.open(fname);	
	myfile << "#\tThis direct azyclic graph was generated by the program generate_dag by Thomas Buechler"  << endl;
	myfile << "#\tThe DAG has " << N << " nodes and " << edges.size() << " edges" << endl;
	myfile << "H\tVN:Z:1.0" << endl;
	for(unsigned int i = 0; i < N; i++)
		myfile << "S\t" << i << "\t" << rb() << endl;	
	for (auto e = edges.begin(); e!=edges.end(); e++){
		unsigned int u = (*e) >> 32;
		unsigned int v = (*e) %  4294967296; //2^32
		myfile << "L\t" << u << "\t+\t" << v << "\t+\t0M" << endl;
	}
	myfile.close();
}

int main(int argc,  char** argv){	
	unsigned int N = 10; // number of nodes
	unsigned int E_min = 30; // number of edges
	unsigned int E_max = 30; // number of edges
	unsigned int P_min = 4; //number of paths
	unsigned int P_max = 4; //number of paths
	unsigned int PL_min = 6; // length of path
	unsigned int PL_max = 6; // length of path
	unsigned int count = 1; //number of graphs to generate
	unsigned int seed = 0; //seed for random_generator
	unsigned int version = 1;
	string f_prefix = ".";	
	
	if(argc <= 1){
		print_help();
		return 0;
	}
	
	try{
		for (int i = 1; i < argc;) {
			string param = argv[i++];
			
			if( param == "-h" || param == "--help"){
				print_help();
				return 0;
			}else if(param == "-v2"){
				version = 2;
				continue;
			}
			
			
			if(i == argc)
				throw invalid_argument("no value for argument "+ param);
			string value = argv[i++];
			
			if(param == "-n"){
				N = stoi(value);
				if(N > 4294967295) throw invalid_argument("N should be smaller than 2^32");
			}else if (param == "-e"){
				E_min = stoi(value);
				E_max = E_min;
			}else if (param == "-e_min"){
				E_min = stoi(value);
			}else if (param == "-e_max"){
				E_max = stoi(value);
			}else if (param == "-p"){
				P_min = stoi(value);
				P_max = P_min;
			}else if (param == "-p_min"){
				P_min = stoi(value);
			}else if (param == "-p_max"){
				P_max = stoi(value);
			}else if (param == "-pl"){
				PL_min = stoi(value);
				PL_max = PL_min;
			}else if (param == "-pl_min"){
				PL_min = stoi(value);
			}else if (param == "-pl_max"){
				PL_max = stoi(value);
			}else if (param == "-c"){
				count = stoi(value);
			}else if (param == "-s"){
				seed = stoi(value);
			}else if (param == "-o"){
				f_prefix = value;
			}else{
				throw invalid_argument("Unknown argument " + param);
			}
		}
	}
	catch(exception e){
		cout << "Error: " << e.what() << endl;
		cout << "type -h or --help for help\n";
	}
	
	my_lkg lkg = my_lkg(seed); //linearer kongruenz generator
	   interval P = interval(P_min,P_max);
	   interval PL = interval(PL_min,PL_max);
	   interval E = interval(E_min,E_max);

	for(unsigned int i = 0; i < count; i++){
		
		set<unsigned long> edges;
		string fname;
		unsigned int s = lkg(); //seed for graph
		
		switch(version){
			case 1: tie(edges, fname) = version_1(N,P,PL,s); break;
			default: tie(edges, fname) = version_2(N,E,s); break;
		}
		
		fname = f_prefix + "/" + fname + ".gfa";
		
		write_gfa_file(fname,N,edges, s);
		
		if(access((fname).c_str(), F_OK))
			throw runtime_error("File " + fname + " was not written correcly. May the directory does not exist.");
	
	}
	
}
